-- Lua Switch Mapping widget by RonRon on RCGroups
-- Lua Switch Mapping widget by RonRon on RCGroups
-- requires at least Ethos 1.1.0 alpha 12

-- Changelog:
--  1.0  Initial version by ronron on RCGroups that allows to predefine text
--       descriptions in the source .lua file.
--  1.1  Minor additions by xStatiCa on RCGroups to add defining the text
--       descriptions on the transmitter when configuring the widget.  The info
--       is saved so that the transmitter will keep the information.  This also
--       allows each model to have it's own configuration.Version 1.1
--  1.2  Fix for reversed LS and RS in some lines
--  1.3  Added option to display switch names so that only the description is configured
--       Added option to display switch lines that have empty discriptions or not
--  1.4  Changes for Ethos 1.1.0 alpha 8  (changed lcd.setFont to lcd.font and lcd.setColor with lcd.color)
--  1.5  Changes for Ethos 1.1.0 alpha 12 (changed storage functions to alpha 12 storage functions)
--  1.6  Moved script into a subfolder and includes image in the new folder so that you don't have to put it in the global bitmaps directory separately
--  1.7  Added check and text to display "Unsupported widget size" if you use a widget other than a large widget.
--  1.8  Changed text for SM to POT3 which is the real name of the slider in the middle bottom of the transmitter
--  1.9  Fixed bug where Middle Slider (Pot3) info was not properly stored/read (by Alfred Vink)
--       Added RIGHT alignment text for RH switches and CENTERED alignment for text for middle slider (POT3) (by Alfred Vink)
--  1.10 Fixed bug where "SE "was shown even if Display Switch Names was off (by Alfred Vink) 
--  1.11 B. Jurisch 01.01.2023 org. swmap Source expanded and adapted so that as many transmitter as possible can be covered
--  1.12 B. Jurisch 03.01.2023 add config switch names if configured and if Titel switch OFF load Transmitter_0.lua 
--  1.13 B. Jurisch 05.01.2023 add overwrite Text color, line1 color, line2, color, and Text Font for switch, and display lau ram over debug_mem flag
--  1.14 B. Jurisch 09.01.2023 add lua_ram in widget set flag over Configuration for runing time
--  1.15 B. Jurisch 27.01.2023 bug in xe.lua POT4 
--  1.16 B. Jurisch 27.01.2023 bug in error messeage for not found switch
--  1.17 B. Jurisch 05.02.2023 show text for switch pos
--  1.18 B. Jurisch 19.02.2023 add TR1-TR4 
--  1.19 B. Jurisch 23.02.2023 error when switching to switch positions in X20 write(widget) ill not save in storage
--  1.20 B. Jurisch 13.03.2023 added default values
--  1.21 B. Jurisch 13.03.2023 remove switsch positions for the moment 
--  1.22 B. Jurisch 14.03.2023 write problem in storage correction 
--  1.23 B. Jurisch 15.03.2023 program optimization tables put together only one default table
--  1.24 B. Jurisch 15.03.2023 copy from other modell file and load defaults if activate
--  1.25 B. Jurisch 16.03.2023 all configuration data will save in file not in memory
--  1.26 B. Jurisch 29.03.2023 mark switches and pots that have been moved, load model after resetting widget, display x and y Positionin Simulator

-- dies sollte probleme lösen wenn zwei scripte die gleichen Funktionen verwenden
--local newEnv = {}
--setmetatable(newEnv, {__index = _ENV})
--local _ENV = newEnv

local version="1.26"
local translations = {en="Switch Mapping", de="Schalter Zuordnung", fr="Telemetrie"}
local debug_flag=false          --> for debug mode

-- Get information for Transmitter
local sys = system.getVersion()   	
local RADIO=string.sub(sys.board,1,3)	  	
local load_flag=false
local sw_pos_status=0   -- 0 unbekant, 1 up, 2 mitte, 3 down

local old_mem={}
local bitmap=nil
local sw_pos=nil
local err_msg=""
local old_trimm=false
local old_fsw=false
local old_h=0

local def_std=nil
local def_erw=nil
local def_pot=nil
local def_fus=nil
local defcol={}

local sw_position={}
local sw_old_position={}
local anz_schalter_pos=0
local sw_tast=""

local ana_pos={}
local ana_anzahl=0
local pos_msg=nil
local sys=nil 
local sys_font_h=0
local titel_off=false
-- Language for display and debug
local langu={}
if system.getLocale() == "de" then	
	langu={"Schalternamen anzeigen",
		   "Hinzufügen Trimm T1-T6",
		   "Schalter ",
		   "Regler ",
		   "Hintergrundfarbe",
		   "Schriftfarbe",
		   "Zeilenfarbe",
		   "Pos. nicht Definiert: ",
		   "Positionen konnten nicht geladen werden.",
		   "Hinzufügen Funktionschalter",
		   "Funktionschalter",
		   "Schalternamen wenn konfiguriert",
		   "LUA RAM anzeigen",
		   "Schalter Position anzeigen",
		   "Kopiere von Modell",
		   "Lade Defaults wenn leer",
		   "Kreis um Schater und Poti",
		  }
	
else
	langu={"Display switch names",
		   "Display Trimm T1-T6",
		   "Switch ",
		   "Slider ",
		   "Backcolor",
		   "Fontcolor",
		   "Linecolor",
		   "Pos. found for: ", 
		   " Positions could not be loaded",
		   "add functionswitch",
		   "functionswitch",
		   "Switch names if configured",
		   "Display LUA RAM",
		   "display switch position",
		   "copy from modell",
		   "load defaults if space",
		   "cirle to switch and sliter",
		  }
end  


-- debug function
function deprint(txt)
	if debug_flag==true then
		print(txt)
	end
end


-- analog position
local function analog_pos()
	--deprint("function analog_pos")
	
	local ok=false
	for i=0,100,1 do 
		local me = system.getSource({category=CATEGORY_ANALOG, member=i})
		if me~=nil then
			if ana_pos[i]==nil then
				ana_pos[i]={} 
				ana_pos[i][1]=me:name()
				ana_pos[i][2]=me:stringValue()
				ana_pos[i][3]=me:stringValue()
			end	
			
			if ana_pos[i][1]==me:name() and ana_pos[i][3]~=me:stringValue() then 
				sw_tast=me:name()
				ana_pos[i][2]=me:stringValue()
				ana_pos[i][3]=me:stringValue()
				ok = true
			end
			if me:name()==nil or me:name()=="---" then ana_anzahl=i-1 break end
		end
	end
	return ok
end

-- search switch position
local function switch_pos()
	--deprint("function switch_pos")
	local ok=false
	for i=0,100,1 do 
		local me = system.getSource({category=CATEGORY_SWITCH_POSITION, member=i})
		if me~=nil then
			if string.sub(me:name(),1,1)=="S" then 
				--deprint ("Schalter Name "..i.." ".. me:name().." sw laenge:"..string.len(me:name()))
				--deprint ("Schalter stringValue "..i.." "..me:stringValue())
				if sw_position[i]==nil then sw_position[i]={} sw_position[i][3]="" end
				if sw_old_position[i]==nil then
					sw_old_position[i]={} 
					sw_old_position[i][1]=me:name()
					sw_old_position[i][2]=me:state()
				end
				sw_position[i][1]=me:name()
				sw_position[i][2]=me:state()
				if sw_position[i][2]~=sw_old_position[i][2] then
					--deprint("Schalter: "..me:name().." position getauscht: ")
					sw_old_position[i][2]=sw_position[i][2]
					sw_tast=string.sub(me:name(),1,2)
					ok=true
				end
			else
				anz_schalter_pos=i-1
				break
			end
		end
	end
	--deprint("ermittelte anz_schalter_pos:"..anz_schalter_pos)
	return ok
	
end

-- search function switch
local function su_log_sw(widget,na)
	--deprint("function su_log_sw")
	local ok = false
	for i=0,widget.anz_fsw,1 do
		if widget.FSW[widget.anz_fsw][0]==na then ok=true break end
	end
	return ok
end

-- add function switch
local function add_logswitch(widget)
	--deprint("LOG Schalter"..widget.anz_fsw)
	if RADIO=="X10" or RADIO=="X12" then widget.anz_fsw=-1 return true end
	if widget.anz_fsw==-1 then
		for y=1,6,1 do
			widget.anz_fsw=widget.anz_fsw+1
			if widget.FSW==nil then widget.FSW={} end
			if widget.FSW[widget.anz_fsw]==nil then widget.FSW[widget.anz_fsw]={}  end
			widget.FSW[widget.anz_fsw][0]="FSCH"..y
			widget.FSW[widget.anz_fsw][1]=""
			--deprint("Fsw hinzugefuegt: "..widget.FSW[widget.anz_fsw][0].." anz:"..widget.anz_fsw)
		end
	end
end

-- function for addition of Trimmslider, TR5 and TR6
local function add_trimm(widget)
	--deprint("Add Trimm")
	if widget.ANALOG[widget.anz_ana][0]~="TR6" then
		--deprint("Add Trimm hinzufügen")
		--TR1
		widget.anz_ana=widget.anz_ana+1
		if widget.ANALOG[widget.anz_ana]==nil then widget.ANALOG[widget.anz_ana]={} end
		widget.ANALOG[widget.anz_ana][0]="TR1"
		widget.ANALOG[widget.anz_ana][1]=""
		--TR2
		widget.anz_ana=widget.anz_ana+1
		if widget.ANALOG[widget.anz_ana]==nil then widget.ANALOG[widget.anz_ana]={} end
		widget.ANALOG[widget.anz_ana][0]="TR2"
		widget.ANALOG[widget.anz_ana][1]=""
		--TR3
		widget.anz_ana=widget.anz_ana+1
		if widget.ANALOG[widget.anz_ana]==nil then widget.ANALOG[widget.anz_ana]={} end
		widget.ANALOG[widget.anz_ana][0]="TR3"
		widget.ANALOG[widget.anz_ana][1]=""
		--TR4
		widget.anz_ana=widget.anz_ana+1
		if widget.ANALOG[widget.anz_ana]==nil then widget.ANALOG[widget.anz_ana]={} end
		widget.ANALOG[widget.anz_ana][0]="TR4"
		widget.ANALOG[widget.anz_ana][1]=""
		--TR5
		widget.anz_ana=widget.anz_ana+1
		if widget.ANALOG[widget.anz_ana]==nil then widget.ANALOG[widget.anz_ana]={} end
		widget.ANALOG[widget.anz_ana][0]="TR5"
		widget.ANALOG[widget.anz_ana][1]=""
		-- TR6
		widget.anz_ana=widget.anz_ana+1
		if widget.ANALOG[widget.anz_ana]==nil then widget.ANALOG[widget.anz_ana]={} end
		widget.ANALOG[widget.anz_ana][0]="TR6"
		widget.ANALOG[widget.anz_ana][1]=""
	end
end

-- search switch
function su_sw(widget,na,su)
	--deprint("search "..na.." in "..su)
	local pos=-1

	if su=="SW1" then 
		if sw_position==nil then return end
		for i=0,anz_schalter_pos,1 do
			--if sw_position[i][1]==na then pos=i --deprint("Fond :"..na.." in SW1") break end
			if sw_position[i][1]==na then pos=i  break end
		end
	end
	
	if su=="SW" then 
		if widget.anz_sw==nil then return end
		for i=0,widget.anz_sw,1 do
			--if widget.SW[i][0]==na then pos=i --deprint("Fond :"..na.." in SW") break end
			if widget.SW[i][0]==na then pos=i break end
		end
	end
	if su=="ANALOG" then 
		if pos==-1 then 
			for i=0,widget.anz_ana,1 do
				--if widget.ANALOG[i][0]==na then pos=i --deprint("Fond :"..na.." in ANALOG") break end
				if widget.ANALOG[i][0]==na then pos=i break end
			end
		end
	end
	if su=="FSW" then 
		if pos==-1 then 
			for i=0,widget.anz_fsw,1 do
				--if widget.FSW[i][0]==na then pos=i --deprint("Fond :"..na.." in FSW") break end
				if widget.FSW[i][0]==na then pos=i break end
			end
		end
	end
	return pos
end

-- search switch position 
local function su_sw_pos(na,su)
	--deprint("function su_sw_pos")
	local pos=-1
	if su=="SW1" then
		na=string.sub(na,1,2)
	end
	for k,v in pairs(sw_pos) do 
		for k1,v1 in pairs(v) do 
			if v1==na then 
			 pos=k
			 break
			end
		end
	end
    return pos
end

-- check if file exist
local function existFile(file)
	--deprint("function existFile:"..file)
	local fileOpen=io.open(file,"r")
	if fileOpen==nil then 
		return false
	else
		local fileClose= io.close(fileOpen)
		return true
	end
end

-- copy file function
local function copyFile(fromfile,tofile)
	--deprint("function copyFile")
	local f=io.open(fromfile,"rb")
	if f==nil then
		return nil, nil
	end
	local content=""
	while true do
		line = io.read(f,"L")
		if (line == nil) then break end
		content=content..line
	end
	io.close(f)
	if content~=nil then
		f=nil
		f = io.open(tofile, "w")
		io.write(f,content)
		io.close(f)
	end
end


local function write_file(file,text)
	--deprint("function write_file")
	text = text.."\n"
	
	if existFile(file)==true then	
		f = io.open(file, "a")
		if f~=nil then 
			io.write(f,text)
			io.close(f)
		else
			local fileOpen=i
			f = io.open(file, "w")
			io.write(f,text)
			io.close(f)
		end
    else
		f = io.open(file, "w")
		io.write(f,text)
		io.close(f)
	end
end

-- write memory
local function write_mem()
	--deprint("function write_mem")
	local mem = system.getMemoryUsage()
	if mem["mainStackAvailable"]~=old_mem[1] or mem["ramAvailable"]~=old_mem[1] or old_mem[2]~=mem["luaRamAvailable"] or old_mem[3]~=mem["luaBitmapsRamAvailable"] then
		local text=""
		old_mem[0]=mem["mainStackAvailable"]
		old_mem[1]=mem["ramAvailable"]
		old_mem[2]=mem["luaRamAvailable"]
		old_mem[3]=mem["luaBitmapsRamAvailable"]
		text=os.date("%d.%m.%Y",os.time())..";"..os.date("%H:%M:%S",os.time())..";"
		if mem["mainStackAvailable"]~=nil then
			text=text..string.format("%.2f",(mem["mainStackAvailable"]/1024))..";"
		else
			text=text..string.format("%.2f",(mem["mainStackAvailable"]))..";"
		end
		
		text=text..string.format("%.2f",(mem["ramAvailable"]/1024))..";"
		text=text..string.format("%.2f",(mem["luaRamAvailable"]/1024))..";"
		text=text..string.format("%.2f",(mem["luaBitmapsRamAvailable"]/1024))..";"
		if mem["luaRamAvailable"]<10 then
			--deprint("MEMORY LEACK:"..mem["luaRamAvailable"]/1024)
			text=text..translations["de"]..";YES;\n"
		else
			text=text..translations["de"]..";NO;\n"
		end
		file="/logs/mem.csv"
		if existFile(file)==true then	
			f = io.open(file, "a")
			if f~=nil then 
				io.write(f,text)
				io.close(f)
			end
		else
			f = io.open(file, "w")
			tt="Radio: SIMULATOR "..sys.board.."\n"
			io.write(f,tt)
			tt="Version:"..sys.version.."\n"
			io.write(f,tt)
			tt="\n"
			io.write(f,tt)
			io.write(f,tt)
			tt="Datum;Zeit;"
			tt=tt.."Gesamtspeicher;"
			tt=tt.."Freier Speicher;"
			tt=tt.."LUA RAM frei;"
			tt=tt.."LUA BMP;"
			tt=tt.."Widget"
			tt=tt.."MEMORY LEACK".."\n"
			io.write(f,tt)
			io.write(f,text)
			io.close(f)
		end
	end
end


-- read all configured data tables  
local function load_data(widget,na)
	--deprint("function load_data")
	-- Read data tabels for model
	if na==nil then na=model.name()..".lua" end
	local data_values, data_switch, data_analog, data_fuswitch = dofile("/scripts/swmap/modelle/"..na)
		
	-- set switches and colors	
	widget.backcolor=tonumber(data_values[2])
	widget.farbe=tonumber(data_values[3])
	widget.linecolor=tonumber(data_values[4])
	widget.DisplaySwitchNames=data_values[5]
	widget.DisplayTrimm=data_values[6]
	widget.DisplayFswitch=data_values[7]
	widget.DisplaySW=data_values[8]
	widget.DisplayPos=data_values[9]
	widget.loaddefault=data_values[10]
	if data_values[11]~=nil then widget.circle_sw=data_values[11] end   -- 28.03.2023 B. Jurisch neu in V 1.26
	
	-- load trimm switch?
	old_trimm=widget.DisplayTrimm
	if widget.DisplayTrimm==true then 
		add_trimm(widget)
	end

	-- load function switch?
	old_fsw=widget.DisplayFswitch
	if widget.DisplayFswitch==true then 
		add_logswitch(widget)
	end
	
	-- fill tables for this model
	-- analog values
	for i=0,widget.anz_ana,1 do 
		widget.ANALOG[i][1]= data_analog[i+1]
		--deprint("Analog :"..i.." "..widget.ANALOG[i][0].." "..widget.ANALOG[i][1])
	end	
	-- function switch values
	for i=0,widget.anz_fsw,1 do 
		widget.FSW[i][1]=data_fuswitch[i+1]
		--deprint("Read Fuswitch:"..i.." "..widget.FSW[i][0].." "..widget.FSW[i][1])
	end	
	
	-- switches values
	if widget.DisplayPos==false then 
		for i=0,widget.anz_sw,1 do 
			if data_switch[i+1]==nil then data_switch[i+1]="" end
			widget.SW[i][1]=data_switch[i+1]
		end
		
	else
		for i=0,anz_schalter_pos,1 do 
			sw_position[i][3]=data_switch[i+1]
		end
	end
	data_def=nil
	data_switch=nil
	data_analog=nil
	data_fuswitch=nil
	widget.modell_load_flag=true
end

-- write configured modell data's
local function write_data(widget)
	--deprint("function write_data")
	-- save all configuration values to file
	local data_def="local def_values={"
	local data_analog="local def_ana={"
	local data_fuswitch="local def_fusw={"
	local data_switch="local def_switch={"
	
	-- save widget switches and colors
	data_def=data_def.."\""..version.."\","
	data_def=data_def.."\""..widget.backcolor.."\","
	data_def=data_def.."\""..widget.farbe.."\","
	data_def=data_def.."\""..widget.linecolor.."\","
	data_def=data_def..tostring(widget.DisplaySwitchNames)..","
	data_def=data_def..tostring(widget.DisplayTrimm)..","
	data_def=data_def..tostring(widget.DisplayFswitch)..","
	data_def=data_def..tostring(widget.DisplaySW)..","
	data_def=data_def..tostring(widget.DisplayPos)..","
	data_def=data_def..tostring(widget.loaddefault)..","
	data_def=data_def..tostring(widget.circle_sw)..",}\n"

	-- Display Trimmslider?
	if widget.DisplayTrimm==true then 
		add_trimm(widget)
		old_trimm=widget.DisplayTrimm
	else
		if widget.DisplayTrimm~=old_trimm then
			--deprint("Delte TR1 - TR6")
			widget.anz_ana=widget.anz_ana-6
			old_trimm=widget.DisplayTrimm
		end
	end
	
	

	-- save analog values
	for i=0,widget.anz_ana,1 do 
		if widget.ANALOG[i][1]==nil then widget.ANALOG[i][1]="" end
		data_analog = data_analog.."\""..widget.ANALOG[i][1].."\""..","
	end
	data_analog=data_analog.."}".."\n"
	if widget.DisplayFswitch==true then
		add_logswitch(widget)
		old_fsw=widget.DisplayFswitch
	else
		if widget.DisplayFswitch~=old_fsw then
			--deprint("Delte Fswitch")
			widget.FSW=nil
			old_trimm=widget.DisplayTrimm
			widget.anz_fsw=-1
		end
	end
	
	-- save function switch values
	for i=0,widget.anz_fsw,1 do 
		if widget.FSW[i][1]==nil then widget.FSW[i][1]="" end
		data_fuswitch=data_fuswitch.."\""..widget.FSW[i][1].."\""..","
	end
	data_fuswitch=data_fuswitch.."}".."\n"
	
	-- switches values	
	if widget.DisplayPos==false   then 
		-- only switch value
		for i=0,widget.anz_sw,1 do 
			if widget.SW[i][1]==nil then widget.SW[i][1]="" end
			data_switch=data_switch.."\""..widget.SW[i][1].."\""..","
		end

	else
		--switches position values
		for i=0,anz_schalter_pos,1 do
			if sw_position[i][3]==nil then sw_position[i][3]="" end
			data_switch=data_switch.."\""..sw_position[i][3].."\""..","
		end
	end
	data_switch=data_switch.."}".."\n".."return def_values, def_switch, def_ana, def_fusw".."\n"
	
	-- write modell data to file
	file="/scripts/swmap/modelle/"..model.name()..".lua"
	--deprint("SChreibe Datei"..file)
	
	f = io.open(file, "w")
	io.write(f,data_def..data_analog..data_fuswitch..data_switch)
	
	io.close(f)
	data_def=nil
	data_switch=nil
	data_analog=nil
	data_fuswitch=nil
end

-- load an other file into table
local function loadLib(name)
  --deprint("laodlib: "..name)
  local lib = dofile("/scripts/swmap/lib/"..name..".lua")
  --if lib==nil then deprint(RADIO.." "..langu[9]) end
  return lib
end

-- initial function for load files
local function initLibs(na)
  --deprint("function 	initLibs")
  if sw_pos==nil then
	sw_pos=loadLib(na)
  end
  -- 15.03.2022 ------------ start------------
  if existFile("/scripts/swmap/lib/default.lua")==true then	
	if def_std==nil then
		def_std, def_erw, def_pot, def_fus=dofile("/scripts/swmap/lib/default.lua")
	end
  end
  -- 15.03.2022 ------------ ende------------
  --if sw_pos == nil then --deprint(RADIO..langu[9]) end

end
function check_display(widget)
	--deprint("function check_display")
	-- if titel off ?
	-- X10 480*272
	-- X12 480*272
	-- X18 480*320
	-- TWN 480*320
	-- X20 800*480
	-- XE  800*480
	--deprint("Check Display Radio:"..RADIO.." w:"..widget.w.." h:"..widget.h)
	titel_off=false

	if (RADIO=="X10" or RADIO=="X12") and widget.h>158  then
		titel_off=true
	end
	if (RADIO=="X18" or RADIO=="TWN") and widget.h>194  then
		titel_off=true
	end
	if (RADIO=="X20" or RADIO=="XE" or RADIO=="XES" ) and widget.h>294  then
		titel_off=true
	end
	
	if titel_off==true and widget.h~=old_h then
		sw_pos=nil
		sw_pos=loadLib(RADIO.."_0")
		old_h=widget.h
		--print("Lade:"..RADIO.."_0")
	else
		if widget.h~=old_h then
			widget.sw_pos=nil
			sw_pos=loadLib(RADIO)
			old_h=widget.h
			--print("Lade:"..RADIO)
		end
	end
	
end

--function for display all switches and slider and lines
function display(widget,na,su)
	--deprint("function display:"..na.." "..su)
	--print("na:"..na.." su:"..su)
	pos=su_sw(widget,na,"SW")  
	
	if pos<0 then pos=su_sw(widget,na,su) end
	
	
	
	idx=su_sw_pos(na,su)
	ok=true
	if idx~=-1 and sw_pos[idx][2]==0 and sw_pos[idx][3]==0 then
		ok=false
	end
	lcd.color(widget.farbe)
	lcd.font(FONT_S)
	local s_font_w, s_font_h = lcd.getTextSize(na)
	s_font_h=s_font_h-2
	
	--print("in Display :"..na.." pos:"..pos.." idx:"..idx.." x:"..sw_pos[idx][2].." Y:"..sw_pos[idx][3]  )
	if pos~=-1 and idx~=-1 and ok==true then 
		local dis_na=""   
		if su=="SW" then dis_na=widget.SW[pos][1] end
		if su=="SW1" then 
			if sw_position[pos][3]==nil then sw_position[pos][3]="" end
			dis_na=sw_position[pos][3] 
		end
		
		if su=="ANALOG" then dis_na=widget.ANALOG[pos][1] end
		if su=="FSW" then dis_na=widget.FSW[pos][1] end
		
		lcd.color(widget.farbe)
		if widget.DisplayEmpty or dis_na ~= "" then
			prefixtext=""
		end
		if widget.DisplaySwitchNames == true then
			if sw_pos[idx][2]<0 then
				prefixtext = dis_na.." "..na 
			else
				prefixtext = na.." "..dis_na
			end
		else
			prefixtext = dis_na
		end
		
		if widget.DisplaySW==true and dis_na=="" then 
			prefixtext=""
		end
		--print("prefixtext: "..prefixtext)
		if prefixtext~="" then
			if sw_pos[idx][9]~=nil then
				lcd.color(sw_pos[idx][9])
			else
				lcd.color(widget.farbe)
			end
			if sw_pos[idx][12]~=nil then
				lcd.font(sw_pos[idx][12])
			else
				lcd.font(FONT_S)
			end		
			pos=sw_pos[idx][2] 
			l_pos=sw_pos[idx][2] 
			if sw_pos[idx][2]<0 then
				pos=widget.w-10
				l_pos=widget.w-10
				w, h = lcd.getTextSize(prefixtext)
				sys_font_h=h
			    pos=pos-w+2
				if su=="SW1" and sw_pos[idx][13]~=nil then lcd.color(sw_pos[idx][13]) end 
				lcd.drawText(pos, sw_pos[idx][3], prefixtext)	
			else
				pos=sw_pos[idx][2]
				if su=="SW1" and sw_pos[idx][13]~=nil then lcd.color(sw_pos[idx][13]) end
				lcd.drawText(pos, sw_pos[idx][3],prefixtext)
			end	
			if prefixtext~="" then
				if sw_pos[idx][10]~=nil then
					lcd.color(sw_pos[idx][10])
				else
					lcd.color(widget.linecolor)
				end

				lcd.drawLine(l_pos ,sw_pos[idx][3]+s_font_h,sw_pos[idx][4],sw_pos[idx][3]+s_font_h)
				if sw_pos[idx][5]~=nil and sw_pos[idx][6]~=nil then
					if sw_pos[idx][11]~=nil then
						lcd.color(sw_pos[idx][11])
					else
						lcd.color(widget.linecolor)
					end
					lcd.drawLine(sw_pos[idx][5],sw_pos[idx][6]+s_font_h, sw_pos[idx][7],sw_pos[idx][8])
				end
				
				-------------------------- display cirle for switch
				if widget.circle_sw==true then 
					if sw_pos[idx][1]==sw_tast and sw_pos[idx][14]~=nil and sw_pos[idx][15]~=nil and sw_pos[idx][16]~=nil then
						lcd.color(sw_pos[idx][16])
						if RADIO=="X18" or RADIO=="X10" or RADIO=="TWX" then
							for y=5,7,1 do
								lcd.drawCircle(sw_pos[idx][14],sw_pos[idx][15],y)
							end
						else
							for y=8,10,1 do
								lcd.drawCircle(sw_pos[idx][14],sw_pos[idx][15],y)
							end
						end
					end	
				end
				--[[ Display Status up, midel or down
				if sw_pos_status==1 then 
					for ix=1,5,1 do
						lcd.drawLine(sw_pos[idx][14]-ix,sw_pos[idx][15]-ix,sw_pos[idx][14]+ix,sw_pos[idx][15]-ix)
					end
				end
				if sw_pos_status==2 then
					for y=1,3,1 do
						lcd.drawCircle(sw_pos[idx][14],sw_pos[idx][15],y)
					end
				end
				if sw_pos_status==3 then 
					print("down")
					for ix=1,5,1 do
						lcd.drawLine(sw_pos[idx][14]+ix,sw_pos[idx][15]+ix+3,sw_pos[idx][14]-ix,sw_pos[idx][15]+ix+3)
					end
				end
				--]]
				
			end
		
		end	
		
	else 
		-- display err for not found swich or slider position
		if idx<0  then
			if err_msg=="" then 
				err_msg=na
			else
				err_msg=err_msg..";"..na
			end
			lcd.color(widget.farbe)
			lcd.drawText(1,widget.h-s_font_h,langu[8]..err_msg)
		end
	end
	if su=="ANALOG" and string.len(sw_tast)>3 then 
		idx=su_sw_pos(sw_tast,"ANALOG")

			--deprint("Poti Markieren: "..sw_tast.." IDX:"..idx.." sw_pos:"..sw_pos[idx][1])
			--[[lcd.color(sw_pos[idx][16])
			if RADIO=="X18" then
				for y=5,7,1 do
					lcd.drawCircle(sw_pos[idx][14],sw_pos[idx][15],y)
				end
			else
				for y=8,10,1 do
					lcd.drawCircle(sw_pos[idx][14],sw_pos[idx][15],y)
				end
			end
			--]]
			lcd.font(FONT_S)
			for ix=0,ana_anzahl,1 do
				if ana_pos[ix][1]==sw_tast then 
					s_font_w, s_font_h = lcd.getTextSize(ana_pos[ix][2])
					s_font_h=s_font_h/4
					lcd.drawText(sw_pos[idx][14]+10,sw_pos[idx][15]-s_font_h,ana_pos[ix][2])
					break
				end
			end
			
	end
	
end



--------------------------------------------------------------------- from here default ETHOS function's ------------------------------------------------------------------

local function name()
	--deprint("function name")
    local locale = system.getLocale()
    return translations[locale] or translations["en"]
end

local function create()
	--deprint("function create")
	initLibs(RADIO)
	local anz_fsw=-1
	local anz_sw=0
	local anz_ana=0
	local SW={}
	local ANA={}
	local FSW={}
	
	switch_pos()
	--switches
	for i=0,30,1 do 
		local me = system.getSource({category=CATEGORY_SWITCH, member=i})
		if me~=nil and me~="--" then
			na=string.sub(me:name(),1,2)
			if na~="--" then
				--deprint("Switch: "..na)   
				if SW[anz_sw]==nil then SW[anz_sw]={} end
				SW[anz_sw][0]=na
				SW[anz_sw][1]=""
				anz_sw=anz_sw+1
			end
		end
	end
	--slider
	for i=4,100,1 do 
		local me = system.getSource({category=CATEGORY_ANALOG, member=i})
		if me~=nil and me~="--" then
			na=me:name()
			if na~="---" then
				--deprint("Slider: "..na)   
				if ANA[anz_ana]==nil then ANA[anz_ana]={} end
				ANA[anz_ana][0]=na
				ANA[anz_ana][1]=""
				anz_ana=anz_ana+1
			end
		end
	end
	-- correction for switch and slider counter	
	anz_sw=anz_sw-1
	anz_ana=anz_ana-1
	
	local backcolor=THEME_DEFAULT_BGCOLOR
	local farbe=ORANGE
	local linecolor=WHITE
	
	return {DisplayEmpty=true,DisplaySwitchNames=true,DisplayTrimm=false,DisplayFswitch=false,DisplaySW=false,DisplayPos=false, lua_ram=false, backcolor=backcolor,farbe=farbe,linecolor=linecolor, SW=SW, ANALOG=ANA,FSW,anz_sw=anz_sw, anz_ana=anz_ana,anz_fsw=anz_fsw,filename=nil,loaddefault=false,modell_load_flag=false,circle_sw=false,source=nil}

end

local function paint(widget)
	if widget.lua_ram==true then 
		write_mem()
	end
	if widget.DisplayPos==true then
		 switch_pos()
		 analog_pos()
	end

	--if load_flag==false then load_data(widget,model.name()..".lua") end
	widget.w, widget.h = lcd.getWindowSize()
	err_msg=""
	check_display(widget)
	if widget.anz_sw==nil then return end
		
	lcd.color(widget.backcolor)
	lcd.drawFilledRectangle(0, 0, widget.w, widget.h)
		
    lcd.drawBitmap(1,10,bitmap,widget.w-10,widget.h-40)   

	---- Display Switches ----------------------------
	if widget.DisplayPos==true then
		xx=nil
		xx={}
		zz=0
		for i=0,anz_schalter_pos,1 do
			----deprint("Display switch: "..sw_position[i][1].."  "..sw_position[i][3])
			--if sw_position[i][2]==true and sw_position[i][3]~="" then
			if sw_position[i][2]==true then
				if sw_position[i][3]==nil then sw_position[i][3]="" end
				test=string.byte(sw_position[i][1],3,3)
				--- 16.03.2023 Schalter Pos ermitteln-----------
				-----start-------------------------------------
				if string.sub(sw_position[i][1],1,2)==sw_tast then 
					sw_pos_status=0
					if test==198 then sw_pos_status=1 end
					if test==045 then sw_pos_status=2 end
					if test==200 then  sw_pos_status=3 end
				end
				-----ende--------------------------------------
				display(widget,sw_position[i][1],"SW1")	
				zz=zz+1
				xx[zz]=string.sub(sw_position[i][1],1,2)
			end
		end
		-- show all switches that are not in the list xx{}
		for y=0,widget.anz_sw,1 do
			ok=false
			for z=1,zz,1 do
				if xx[z]==widget.SW[y][0] then ok=true break end
			end 
			if ok==false then
				display(widget,widget.SW[y][0],"SW")
			end
		end
	else
		----Display switches------
		for i=0,widget.anz_sw,1 do
			display(widget,widget.SW[i][0],"SW")
		end
	end

	----Display slider------
	for i=0,widget.anz_ana,1 do
		----deprint("display slider: "..widget.ANALOG[i][0])
		display(widget,widget.ANALOG[i][0],"ANALOG")
	end
	----Display Fswitch-----
	for i=0,widget.anz_fsw,1 do
		--deprint("display fsw: "..widget.FSW[i][0])
		display(widget,widget.FSW[i][0],"FSW")
	end
	
	if widget.lua_ram==true then
		left, hight = lcd.getTextSize("LUA RAM:"..(string.format("%.2f",old_mem[2]/1024)).." KB")
		left = left + 10
		if old_mem[2]>0 then 
			lcd.color(ORANGE)
			
			lcd.drawText(widget.w - left,0,"LUA RAM:"..(string.format("%.2f",old_mem[2]/1024)).." KB")
		else
			lcd.color(RED)
			lcd.drawText(widget.w - left,20,"LUA RAM:"..(string.format("%.2f",old_mem[2]/1024)).." KB")
		end
		
	end
	if sys.simulation==true and pos_msg~=nil and pos_msg~="" then 
		lcd.color(WHITE)
		lcd.drawText(5,widget.h-15,pos_msg)
	else 
		pos_msg=nil
	end
end

local function configure(widget)
	--deprint("function configure")
	err_msg=""
	-- Version
	line = form.addLine("swmap Version: "..version) 
	form.addStaticText(line, nil, "")
	
	
	line = form.addLine(langu[16])
	form.addBooleanField(line, nil, function() return widget.loaddefault end, function(value) widget.loaddefault = value end)
	
	if widget.filename==nil then widget.filename="" end
	line=form.addLine(langu[15])
	form.addFileField(line, nil, "/scripts/swmap/modelle/", "any", function() return widget.filename end, function(newValue) widget.filename = newValue end)
	
	-- display switch names
	line = form.addLine(langu[1])
	form.addBooleanField(line, nil, function() return widget.DisplaySwitchNames end, function(value) widget.DisplaySwitchNames = value end)
	-- display switch names if configured
	line = form.addLine(langu[12])
	form.addBooleanField(line, nil, function() return widget.DisplaySW end, function(value) widget.DisplaySW = value end)
	
	-- Display Switch Pos 
	line = form.addLine(langu[14])
	form.addBooleanField(line, nil, function() return widget.DisplayPos end, function(value) widget.DisplayPos = value end)
		
	-- Funktion/logical circle switch
	line = form.addLine(langu[17])
	form.addBooleanField(line, nil, function() return widget.circle_sw end, function(value) widget.circle_sw = value end)	
	
	-- Add Trimm 
	line = form.addLine(langu[2])
	form.addBooleanField(line, nil, function() return widget.DisplayTrimm end, function(value) widget.DisplayTrimm = value end)
	-- Add Funktion/logical switch
	line = form.addLine(langu[10])
	form.addBooleanField(line, nil, function() return widget.DisplayFswitch end, function(value) widget.DisplayFswitch = value end)
	-- Display LUA RAM
	line = form.addLine(langu[13])
	form.addBooleanField(line, nil, function() return widget.lua_ram end, function(value) widget.lua_ram = value end)
	
	
	line = form.addLine(langu[5])
	widget.field_color = form.addColorField(line, nil, function() return widget.backcolor end, function(color) widget.backcolor = color end)
	line = form.addLine(langu[6])
	widget.field_color = form.addColorField(line, nil, function() return widget.farbe end, function(color) widget.farbe = color end)
	line = form.addLine(langu[7])
	widget.field_color = form.addColorField(line, nil, function() return widget.linecolor end, function(color) widget.linecolor = color end)
	
	--Switches with position
	if widget.DisplayPos==true then
		for i=0,anz_schalter_pos,1 do
			-- 14.03.2023 ---- start---------------------
			if sw_position[i][3]==nil then sw_position[i][3]="" end  -- 14.03.2023
			if sw_position[i][3]=="" then 
				if def_erw~=nil and widget.loaddefault==true then 
					for idx=1,anz_schalter_pos+1,1 do 
						if def_erw[idx]~=nil and def_erw[idx][1]==i  then 
							sw_position[i][3]=def_erw[idx][2]
							break;
						end	
					end
				end
			end
			-- 14.03.2023 ---- ende---------------------
			line = form.addLine(sw_position[i][1])
	
			form.addTextField(line, nil, function() return sw_position[i][3] end, function(value) sw_position[i][3] = value end)
		end
	else
		--Switches
		for i=0,widget.anz_sw,1 do 
			-- 14.03.2023 ---- start---------------------
			if widget.SW[i][1]==nil then widget.SW[i][1]="" end 
			if widget.SW[i][1]=="" then 
				if def_std~=nil and widget.loaddefault==true then
					for idx=1,widget.anz_sw+1,1 do 
						if def_std[idx]~=nil and def_std[idx][1]==i  then 
							widget.SW[i][1]=def_std[idx][2]
							break;
						end	
					end
				end
			end
			-- 14.03.2023 ---- ende---------------------
			line = form.addLine(langu[3]..widget.SW[i][0])
			form.addTextField(line, nil, function() return widget.SW[i][1] end, function(value) widget.SW[i][1] = value end)
		end
	end
	
	-- Slider
	for i=0,widget.anz_ana,1 do 
		if widget.ANALOG[i][1]==nil then widget.ANALOG[i][1]="" end
		-- 14.03.2023 ---- start---------------------
		if widget.ANALOG[i][1]=="" then 
			if def_pot~=nil and widget.loaddefault==true then
				for idx=1,widget.anz_ana+1,1 do 
					if def_pot[idx]~=nil and def_pot[idx][1]==i  then 
						widget.ANALOG[i][1]=def_pot[idx][2]
						break;
					end	
				end
			end
		end
		-- 14.03.2023 ---- ende---------------------
		line = form.addLine(langu[4]..widget.ANALOG[i][0])
		form.addTextField(line, nil, function() return widget.ANALOG[i][1] end, function(value) widget.ANALOG[i][1] = value end)
	end
	
	-- Functionswitch
	for i=0,widget.anz_fsw,1 do 
		if widget.FSW[i][1]==nil then widget.FSW[i][1]="" end
		-- 14.03.2023 ---- start---------------------
		if widget.FSW[i][1]=="" then 
			if def_fus~=nil and widget.loaddefault==true then
				for idx=1,widget.anz_fsw+1,1 do 
					if def_fus[idx]~=nil and def_fus[idx][1]==i  then 
						widget.FSW[i][1]=def_fus[idx][2]
						break;
					end	
				end
			end
		end
		-- 14.03.2023 ---- ende---------------------
		line = form.addLine(langu[11]..widget.FSW[i][0])
		form.addTextField(line, nil, function() return widget.FSW[i][1] end, function(value) widget.FSW[i][1] = value end)
	end
	
	
end



local function write(widget)
	--deprint("function write")
	if widget.filename~=nil and string.len(widget.filename)>=1 then 
		--deprint("lade 1")
		load_data(widget,widget.filename..".lua") 
		widget.filename=nil
	end
	if widget.modell_load_flag==false  then 
		if existFile("/scripts/swmap/modelle/"..model.name()..".lua" )==true then 
  		   --deprint("lade 2")
			widget.modell_load_flag=true
			load_data(widget,model.name()..".lua" ) 
		end
	else
		--deprint("schreibe")
		load_flag=false
		write_data(widget)
	end
end

local function wakeup(widget)
	    lcd.invalidate()
end
local function event(widget, category, value, x, y)
	--deprint("Event received:", category, value, "x:"..x, "Y:"..y)
	if (sys.simulation==true and category==EVT_TOUCH and y<5) or (sys.simulation==true and category==0 and value==35) then 
		--print("LOAD")
		system.killEvents(value)
		old_h=0
		check_display(widget)
		return true 
	end
	if (sys.simulation==true and category==EVT_TOUCH)  then 
		system.killEvents(value)
		if titel_off==false then 
			if RADIO=="X20" then
				y1=y-37	
				y2=y-22
				y3=y-19
			end
			if RADIO=="XES" or RADIO=="X18"  then
				y1=y-38
				y2=y-22
				y3=y-20
			end
			if RADIO=="TWX" then
				y1=y-11
				y2=y
				y3=y
			end
			if y1<0 then y1=0 end
			pos_msg=RADIO..".lua Position: font:"..y1.."  Line1 X:"..x.." Y:"..y1.." Line 2 X:"..x.." Y:"..y2.." Kreis x:"..x.." y:"..y3
		else
			if RADIO=="X20" then
				y1=y-16
				y2=y
				y3=y-2
				x1=x+15
				x2=x+18
			end	
			if RADIO=="XES" or RADIO =="X18"   then
				y1=y-16
				y2=y-2
				y3=y
			end
			if RADIO=="TWX" then
				y1=y-16
				y2=y-15
				y3=y
			end
			
			if y1<0 then y1=0 end
			 pos_msg=RADIO.."_0.lua Position: font:"..y1.."  Line1 X:"..x.." Y:"..y1.." Line 2 X:"..x.." Y:"..y2.." Kreis x:"..x.." y:"..y3
		end
		return true
	end
end


local function init()
	if bitmap==nil then
		sys = system.getVersion()   
		RADIO=string.sub(sys.board,1,3)	
		--deprint("Radio:"..RADIO)
		file="/scripts/swmap/lib/"..RADIO..".lua"
		if existFile(file)==false then 
			fromfile="/scripts/swmap/lib/x18.lua"
			copyFile(fromfile,file)
		end
		file="/scripts/swmap/bitmaps/"..RADIO..".png"
		if existFile(file)==true then 
			bitmap = lcd.loadBitmap(file)	
		else
			fromfile="/scripts/swmap/bitmaps/x18.png"
			copyFile(fromfile,file)
			bitmap = lcd.loadBitmap(file)	
		end
	end

	system.registerWidget({key="swmap", name=name, create=create, paint=paint,  event=event, wakeup=wakeup,   configure=configure, read=load_data, write=write})

end

return {init=init}
